class RioDeprecationWarning(Warning):
    """
    The user used/did something that's deprecated.
    """


class RioPotentialMistakeWarning(Warning):
    """
    The user did something sketchy that's probably a bad idea.
    """
