logo = "rio/logo"
logo_color = "rio/logo:color"
logo_fill = "rio/logo:fill"
logo_and_text_horizontal = "rio/logo-and-text-horizontal"
logo_and_text_horizontal_color = "rio/logo-and-text-horizontal:color"
logo_and_text_horizontal_fill = "rio/logo-and-text-horizontal:fill"
logo_and_text_vertical = "rio/logo-and-text-vertical"
logo_and_text_vertical_color = "rio/logo-and-text-vertical:color"
logo_and_text_vertical_fill = "rio/logo-and-text-vertical:fill"
