corner_round_bottom_left = "styling/corner-round-bottom-left"
corner_round_bottom_right = "styling/corner-round-bottom-right"
corner_round_top_left = "styling/corner-round-top-left"
corner_round_top_right = "styling/corner-round-top-right"
