from .fundamental_component import FundamentalComponent

__all__ = [
    "DevToolsConnector",
]


class DevToolsConnector(FundamentalComponent):
    pass


DevToolsConnector._unique_id = "DevToolsConnector-builtin"
